=begin
=class NumRu::GPhys -- Fast Fourier transformation and its applications

methods of NumRu::GPhys defined in gphys_fft.rb

=methods
---fft(backward=false, *dims)
    Fast Fourier Transformation (FFT) by using  
    (((<FFTW|URL:http://www.fftw.org>))) ver 3 or ver 2.
    A FFTW ver.2 interface is included in NArray, while
    to use FFTW ver.3, you have to install separately.
    Dimension specification by the argument dims is available
    only with ver.3. By default, FFT is applied to all dimensions.

    The transformation is complex. If the input data is not complex,
    it will be coerced to complex before transformation.

    When the FT is forward, the result is normalized 
    (i.e., divided by the data number), unlike the default behavior of
    FFTW.
    
    Each coordinate is assumed to be equally spaced without checking.
    The new coordinate variables will be set equal to wavenumbers,
    derived as 2*PI/(length of the axis)*[0,1,2,..], where the length
    of the axis is derieved as (coord.val.max - coord.val.min)*(n+1)/n.

    ARGUMENTS
    * backward (true of false) : when true, backward FT is done;
      otherwise forward FT is done.
    * dims (integers) : dimensions to apply FFT

    RETURN VALUE
    * a GPhys

    EXAMPLE
      gphy.fft           # forward, for all dimensions
      gphy.fft(true)     # backward, for all dimensions
      gphy.fft(nil, 0,1) # forward, for the first and second dimensions.
      gphy.fft(true, -1) # backward, for the last dimension.

=end

begin
  require "numru/fftw3"
rescue LoadError
end

module NumRu
  class GPhys
    @@fft_forward = -1
    @@fft_backward = 1


    def fft(backward=false, *dims)
      fftw3 = false
      if defined?(FFTW3)
	fftw3 = true
      elsif !defined?(FFTW)
	raise "Both FFTW3 and FFTW are not installed."
      end
      if backward==true
	dir = @@fft_backward
      elsif !backward
	dir = @@fft_forward
      else
	raise ArgumentError,"1st arg must be true or false (or, equivalenty, nil)"
      end

      # <FFT for all dimensions>
      sp = self.copy  # make a deep clone
      if fftw3
	fcoef = FFTW3.fft( sp.data.val, dir, *dims )
      else
	if dims.length == 0
	  raise ArgumentError,
	    "dimension specification is available only if FFTW3 is installed"
	end
	fcoef = FFTW.fftw( sp.data.val, dir )
      end
      if dir == @@fft_forward
	if dims.length == 0
	  fcoef = fcoef / fcoef.length       # normalized if forward FT
	else
	  sh = fcoef.shape
	  len = 1
	  dims.each{|d|
	    raise ArgumentError, "dimension out of range" if sh[d] == nil
	    len *= sh[d]
          }
	  fcoef = fcoef / len
        end
      end
      sp.data.replace_val( fcoef )

      # <coordinate variables>
      for i in 0...sp.rank
	if dims.length == 0 || dims.include?(i) || dims.include?(i+rank)
	  cv = sp.coord(i).val
	  n = cv.length
	  clen = (cv.max - cv.min) * (n+1) / n
	  wn = (2*Math::PI/clen) * NArray.float(cv.length).indgen!
	  sp.coord(i).replace_val(wn)
	  ( ln = sp.coord(i).get_att('long_name') ) &&
	    sp.coord(i).set_att('long_name','wavenumber - '+ln) 
	  ( un = sp.coord(i).get_att('units') ) &&
	    sp.coord(i).set_att('units','('+un+')-1')
          # ^^developper's memo: Use Toyoda's Units when he has completed it.
	end
      end

      # <fini>
      sp
    end

  end
end
